@@ -8,27 +8,29 @@ import os
 import re
 import sys
 from collections import Counter
-from typing import TYPE_CHECKING
+from typing import Any, Dict, List
 
 from pyxform import aliases, constants
-from pyxform.constants import EXTERNAL_INSTANCE_EXTENSIONS, ROW_FORMAT_STRING
+from pyxform.constants import (
+    _MSG_SUPPRESS_SPELLING,
+    EXTERNAL_INSTANCE_EXTENSIONS,
+    ROW_FORMAT_STRING,
+    XML_IDENTIFIER_ERROR_MESSAGE,
+)
+from pyxform.entities.entities_parsing import (
+    get_entity_declaration,
+    validate_entity_saveto,
+)
 from pyxform.errors import PyXFormError
-from pyxform.utils import default_is_dynamic, is_valid_xml_tag, levenshtein_distance
-from pyxform.validators.pyxform import select_from_file_params
+from pyxform.utils import default_is_dynamic
+from pyxform.validators.pyxform import parameters_generic, select_from_file_params
 from pyxform.validators.pyxform.missing_translations_check import (
     missing_translations_check,
 )
 from pyxform.xls2json_backends import csv_to_dict, xls_to_dict, xlsx_to_dict
-
-if TYPE_CHECKING:
-    from typing import Any, Dict, KeysView, List, Optional
+from pyxform.xlsparseutils import find_sheet_misspellings, is_valid_xml_tag
 
 SMART_QUOTES = {"\u2018": "'", "\u2019": "'", "\u201c": '"', "\u201d": '"'}
-_MSG_SUPPRESS_SPELLING = (
-    " If you do not mean to include a sheet, to suppress this message, "
-    "prefix the sheet name with an underscore. For example 'setting' "
-    "becomes '_setting'."
-)
 
 
 def print_pyobj_to_json(pyobj, path=None):
@@ -275,14 +277,18 @@ def add_flat_annotations(prompt_list, parent_relevant="", name_prefix=""):
                 #    prompt['name'] = name_prefix + prompt['name']
 
 
-def process_range_question_type(row):
-    """Returns a new row that includes the Range parameters start, end and
-    step.
+def process_range_question_type(
+    row: Dict[str, Any], parameters: parameters_generic.PARAMETERS_TYPE
+) -> Dict[str, Any]:
+    """
+    Returns a new row that includes the Range parameters start, end and step.
 
     Raises PyXFormError when invalid range parameters are used.
     """
     new_dict = row.copy()
-    parameters = get_parameters(new_dict.get("parameters", ""), ["start", "end", "step"])
+    parameters = parameters_generic.validate(
+        parameters=parameters, allowed=("start", "end", "step")
+    )
     parameters_map = {"start": "start", "end": "end", "step": "step"}
     defaults = {"start": "1", "end": "10", "step": "1"}
 
@@ -316,34 +322,6 @@ def process_image_default(default_value):
     return default_value
 
 
-def find_sheet_misspellings(key: str, keys: "KeysView") -> "Optional[str]":
-    """
-    Find possible sheet name misspellings to warn the user about.
-
-    It's possible that this will warn about sheet names for sheets that have
-    auxilliary metadata that is not meant for processing by pyxform. For
-    example the "osm" sheet name may be similar to many other initialisms.
-
-    :param key: The sheet name to look for.
-    :param keys: The workbook sheet names.
-    """
-    candidates = tuple(
-        _k  # thanks to black
-        for _k in keys
-        if 2 >= levenshtein_distance(_k.lower(), key)
-        and _k not in constants.SUPPORTED_SHEET_NAMES
-        and not _k.startswith("_")
-    )
-    if 0 < len(candidates):
-        msg = (
-            "When looking for a sheet named '{k}', the following sheets with "
-            "similar names were found: {c}."
-        ).format(k=key, c=str(", ".join(("'{}'".format(c) for c in candidates))))
-        return msg
-    else:
-        return None
-
-
 def workbook_to_json(
     workbook_dict,
     form_name=None,
@@ -559,6 +537,9 @@ def workbook_to_json(
                                 )
                             )  # noqa
 
+    # ########## Entities sheet ###########
+    entity_declaration = get_entity_declaration(workbook_dict, warnings)
+
     # ########## Survey sheet ###########
     survey_sheet = workbook_dict[constants.SURVEY]
     # Process the headers:
@@ -669,6 +650,8 @@ def workbook_to_json(
                 ROW_FORMAT_STRING % row_number + " Question with no type.\n" + str(row)
             )
 
+        parameters = parameters_generic.parse(raw_parameters=row.get("parameters", ""))
+
         # Pull out questions that will go in meta block
         if question_type == "audit":
             # Force audit name to always be "audit" to follow XForms spec
@@ -681,16 +664,16 @@ def workbook_to_json(
             row["name"] = "audit"
 
             new_dict = row.copy()
-            parameters = get_parameters(
-                new_dict.get("parameters", ""),
-                [
+            parameters_generic.validate(
+                parameters=parameters,
+                allowed=(
                     constants.LOCATION_PRIORITY,
                     constants.LOCATION_MIN_INTERVAL,
                     constants.LOCATION_MAX_AGE,
                     constants.TRACK_CHANGES,
                     constants.IDENTIFY_USER,
                     constants.TRACK_CHANGES_REASONS,
-                ],
+                ),
             )
 
             if constants.TRACK_CHANGES in parameters.keys():
@@ -898,13 +881,12 @@ def workbook_to_json(
         if not is_valid_xml_tag(question_name):
             if isinstance(question_name, bytes):
                 question_name = question_name.encode("utf-8")
-            error_message = ROW_FORMAT_STRING % row_number
-            error_message += " Invalid question name [" + question_name + "] "
-            error_message += "Names must begin with a letter, colon," + " or underscore."
-            error_message += (
-                "Subsequent characters can include numbers," + " dashes, and periods."
+
+            raise PyXFormError(
+                f"{ROW_FORMAT_STRING % row_number} Invalid question name '{question_name}'. Names {XML_IDENTIFIER_ERROR_MESSAGE}"
             )
-            raise PyXFormError(error_message)
+
+        validate_entity_saveto(row, row_number, entity_declaration)
 
         # Try to parse question as begin control statement
         # (i.e. begin loop/repeat/group):
@@ -1123,8 +1105,8 @@ def workbook_to_json(
                     select_params_allowed += ["value", "label"]
 
                 # Look at parameters column for select parameters
-                parameters = get_parameters(
-                    row.get("parameters", ""), select_params_allowed
+                parameters_generic.validate(
+                    parameters=parameters, allowed=select_params_allowed
                 )
 
                 if "randomize" in parameters.keys():
@@ -1240,7 +1222,7 @@ def workbook_to_json(
 
         # range question_type
         if question_type == "range":
-            new_dict = process_range_question_type(row)
+            new_dict = process_range_question_type(row=row, parameters=parameters)
             parent_children_array.append(new_dict)
             continue
 
@@ -1249,8 +1231,7 @@ def workbook_to_json(
 
             if row.get("default"):
                 new_dict["default"] = process_image_default(row["default"])
-            # Validate max-pixels
-            parameters = get_parameters(row.get("parameters", ""), ["max-pixels"])
+            parameters_generic.validate(parameters=parameters, allowed=("max-pixels",))
             if "max-pixels" in parameters.keys():
                 try:
                     int(parameters["max-pixels"])
@@ -1269,7 +1250,7 @@ def workbook_to_json(
 
         if question_type == "audio":
             new_dict = row.copy()
-            parameters = get_parameters(row.get("parameters", ""), ["quality"])
+            parameters_generic.validate(parameters=parameters, allowed=("quality",))
 
             if "quality" in parameters.keys():
                 if parameters["quality"] not in [
@@ -1288,7 +1269,7 @@ def workbook_to_json(
 
         if question_type == "background-audio":
             new_dict = row.copy()
-            parameters = get_parameters(row.get("parameters", ""), ["quality"])
+            parameters_generic.validate(parameters=parameters, allowed=("quality",))
 
             if "quality" in parameters.keys():
                 if parameters["quality"] not in [
@@ -1308,13 +1289,17 @@ def workbook_to_json(
             new_dict = row.copy()
 
             if question_type == "geopoint":
-                parameters = get_parameters(
-                    row.get("parameters", ""),
-                    ["allow-mock-accuracy", "capture-accuracy", "warning-accuracy"],
+                parameters_generic.validate(
+                    parameters=parameters,
+                    allowed=(
+                        "allow-mock-accuracy",
+                        "capture-accuracy",
+                        "warning-accuracy",
+                    ),
                 )
             else:
-                parameters = get_parameters(
-                    row.get("parameters", ""), ["allow-mock-accuracy"]
+                parameters_generic.validate(
+                    parameters=parameters, allowed=("allow-mock-accuracy",)
                 )
 
             if "allow-mock-accuracy" in parameters.keys():
@@ -1351,7 +1336,6 @@ def workbook_to_json(
 
             parent_children_array.append(new_dict)
             continue
-
         # TODO: Consider adding some question_type validation here.
 
         # Put the row in the json dict as is:
@@ -1395,6 +1379,10 @@ def workbook_to_json(
             }
         )
 
+    if len(entity_declaration) > 0:
+        json_dict[constants.ENTITY_RELATED] = "true"
+        meta_children.append(entity_declaration)
+
     if len(meta_children) > 0:
         meta_element = {
             "name": "meta",
@@ -1405,7 +1393,7 @@ def workbook_to_json(
         survey_children_array = stack[0]["parent_children"]
         survey_children_array.append(meta_element)
 
-    # print_pyobj_to_json(json_dict)
+    print_pyobj_to_json(json_dict)
     return json_dict
 
 
@@ -1480,34 +1468,6 @@ def organize_by_values(dict_list, key):
     return result
 
 
-def get_parameters(raw_parameters, allowed_parameters):
-    parts = raw_parameters.split(";")
-    if len(parts) == 1:
-        parts = raw_parameters.split(",")
-    if len(parts) == 1:
-        parts = raw_parameters.split()
-
-    params = {}
-    for param in parts:
-        if "=" not in param:
-            raise PyXFormError(
-                "Expecting parameters to be in the form of "
-                "'parameter1=value parameter2=value'."
-            )
-        k, v = param.split("=")[:2]
-        key = k.lower().strip()
-        if key in allowed_parameters:
-            params[key] = v.lower().strip()
-        else:
-            raise PyXFormError(
-                "Accepted parameters are "
-                "'%s': '%s' is an invalid parameter."
-                % (", ".join(allowed_parameters), key)
-            )
-
-    return params
-
-
 class SpreadsheetReader:
     def __init__(self, path_or_file):
         path = path_or_file
